@@ -38,9 +38,6 @@ extern "C" {
 
 #define GRN_N_RESERVED_TYPES 256
 
-#define GRN_JSON_LOAD_OPEN_BRACKET 0x40000000
-#define GRN_JSON_LOAD_OPEN_BRACE   0x40000001
-
 typedef struct _grn_db grn_db;
 typedef struct _grn_proc grn_proc;
 
@@ -65,6 +62,61 @@ typedef struct {
   grn_id range;
 } grn_obj_spec;
 
+grn_bool grn_db_spec_unpack(grn_ctx *ctx,
+                            grn_id id,
+                            void *encoded_spec,
+                            uint32_t encoded_spec_size,
+                            grn_obj_spec **spec,
+                            grn_obj *decoded_spec,
+                            const char *error_message_tag);
+
+#define GRN_DB_SPEC_EACH_BEGIN(ctx, cursor, id, spec) do {              \
+  grn_obj *db = grn_ctx_db((ctx));                                      \
+  grn_db *db_raw = (grn_db *)db;                                        \
+  grn_obj decoded_spec;                                                 \
+  grn_io_win iw;                                                        \
+  grn_bool iw_need_unref = GRN_FALSE;                                   \
+  GRN_OBJ_INIT(&decoded_spec, GRN_VECTOR, 0, GRN_DB_TEXT);              \
+  GRN_TABLE_EACH_BEGIN((ctx), db, cursor, id) {                         \
+    void *encoded_spec;                                                 \
+    uint32_t encoded_spec_size;                                         \
+    grn_bool success;                                                   \
+    grn_obj_spec *spec;                                                 \
+                                                                        \
+    if (iw_need_unref) {                                                \
+      grn_ja_unref(ctx, &iw);                                           \
+      iw_need_unref = GRN_FALSE;                                        \
+    }                                                                   \
+    encoded_spec = grn_ja_ref((ctx),                                    \
+                              db_raw->specs,                            \
+                              id,                                       \
+                              &iw,                                      \
+                              &encoded_spec_size);                      \
+    if (!encoded_spec) {                                                \
+      continue;                                                         \
+    }                                                                   \
+    iw_need_unref = GRN_TRUE;                                           \
+                                                                        \
+    GRN_BULK_REWIND(&decoded_spec);                                     \
+    success = grn_db_spec_unpack(ctx,                                   \
+                                 id,                                    \
+                                 encoded_spec,                          \
+                                 encoded_spec_size,                     \
+                                 &spec,                                 \
+                                 &decoded_spec,                         \
+                                 __FUNCTION__);                         \
+   if (!success) {                                                      \
+     continue;                                                          \
+   }                                                                    \
+
+#define GRN_DB_SPEC_EACH_END(ctx, cursor)         \
+  } GRN_TABLE_EACH_END(ctx, cursor);              \
+  if (iw_need_unref) {                            \
+    grn_ja_unref(ctx, &iw);                       \
+  }                                               \
+  GRN_OBJ_FIN((ctx), &decoded_spec);              \
+} while(GRN_FALSE)
+
 void grn_db_init_from_env(void);
 
 GRN_API grn_rc grn_db_close(grn_ctx *ctx, grn_obj *db);
@@ -83,6 +135,10 @@ grn_id grn_table_get_v(grn_ctx *ctx, grn_obj *table, const void *key, int key_si
                        void **value);
 grn_id grn_table_add_v(grn_ctx *ctx, grn_obj *table, const void *key, int key_size,
                        void **value, int *added);
+grn_id grn_table_add_by_key(grn_ctx *ctx,
+                            grn_obj *table,
+                            grn_obj *key,
+                            int *added);
 GRN_API grn_rc grn_table_get_info(grn_ctx *ctx, grn_obj *table, grn_table_flags *flags,
                                   grn_encoding *encoding, grn_obj **tokenizer,
                                   grn_obj **normalizer,
@@ -404,38 +460,11 @@ void grn_obj_spec_save(grn_ctx *ctx, grn_db_obj *obj);
 
 grn_rc grn_obj_reinit_for(grn_ctx *ctx, grn_obj *obj, grn_obj *domain_obj);
 
-#define GRN_INT32_POP(obj,value) do {\
-  if (GRN_BULK_VSIZE(obj) >= sizeof(int32_t)) {\
-    GRN_BULK_INCR_LEN((obj), -(sizeof(int32_t)));\
-    value = *(int32_t *)(GRN_BULK_CURR(obj));\
-  } else {\
-    value = 0;\
-  }\
-} while (0)
-
-#define GRN_UINT32_POP(obj,value) do {\
-  if (GRN_BULK_VSIZE(obj) >= sizeof(uint32_t)) {\
-    GRN_BULK_INCR_LEN((obj), -(sizeof(uint32_t)));\
-    value = *(uint32_t *)(GRN_BULK_CURR(obj));\
-  } else {\
-    value = 0;\
-  }\
-} while (0)
-
 void grn_expr_pack(grn_ctx *ctx, grn_obj *buf, grn_obj *expr);
 GRN_API grn_rc grn_expr_inspect(grn_ctx *ctx, grn_obj *buf, grn_obj *expr);
 grn_hash *grn_expr_get_vars(grn_ctx *ctx, grn_obj *expr, unsigned int *nvars);
 grn_obj *grn_expr_open(grn_ctx *ctx, grn_obj_spec *spec, const uint8_t *p, const uint8_t *pe);
 
-GRN_API void grn_load_(grn_ctx *ctx, grn_content_type input_type,
-                       const char *table, unsigned int table_len,
-                       const char *columns, unsigned int columns_len,
-                       const char *values, unsigned int values_len,
-                       const char *ifexists, unsigned int ifexists_len,
-                       const char *each, unsigned int each_len,
-                       grn_obj *output_ids,
-                       uint32_t emit_level);
-
 GRN_API grn_rc grn_table_group_with_range_gap(grn_ctx *ctx, grn_obj *table,
                                               grn_table_sort_key *group_key,
                                               grn_obj *result_set,
