@@ -1464,24 +1464,29 @@ grn_atoui(const char *nptr, const char *end, const char **rest)
 int64_t
 grn_atoll(const char *nptr, const char *end, const char **rest)
 {
-  /* FIXME: INT_MIN is not supported */
   const char *p = nptr;
-  int n = 0, o = 0;
-  int64_t v = 0, t;
+  int o = 0;
+  int64_t v = 0;
   if (p < end && *p == '-') {
     p++;
-    n = 1;
     o = 1;
-  }
-  while (p < end && *p >= '0' && *p <= '9') {
-    t = v * 10 + (*p - '0');
-    if (t < v) { v = 0; break; }
-    v = t;
-    o = 0;
-    p++;
+    while (p < end && *p >= '0' && *p <= '9') {
+      int64_t t = v * 10 - (*p - '0');
+      if (t > v) { v = 0; break; }
+      v = t;
+      o = 0;
+      p++;
+    }
+  } else {
+    while (p < end && *p >= '0' && *p <= '9') {
+      int64_t t = v * 10 + (*p - '0');
+      if (t < v) { v = 0; break; }
+      v = t;
+      p++;
+    }
   }
   if (rest) { *rest = o ? nptr : p; }
-  return n ? -v : v;
+  return v;
 }
 
 uint64_t
@@ -1760,18 +1765,29 @@ grn_aton(grn_ctx *ctx, const char *p, const char *end, const char **rest,
           grn_obj_reinit(ctx, res, GRN_DB_INT64, 0);
           GRN_INT64_SET(ctx, res, int64);
         }
-      } else if (rest_char == '.' || rest_char == 'e' || rest_char == 'E' ||
-                 (rest_char >= '0' && rest_char <= '9')) {
-        char *rest_float;
-        double d;
-        errno = 0;
-        d = strtod(p, &rest_float);
-        if (!errno && rest_float == end) {
-          grn_obj_reinit(ctx, res, GRN_DB_FLOAT, 0);
-          GRN_FLOAT_SET(ctx, res, d);
-          *rest = rest_float;
-        } else {
-          return GRN_INVALID_ARGUMENT;
+      } else {
+        if (*p != '-' && rest_char >= '0' && rest_char <= '9') {
+          uint64_t uint64 = grn_atoull(p, end, rest);
+          if (end == *rest) {
+            grn_obj_reinit(ctx, res, GRN_DB_UINT64, 0);
+            GRN_UINT64_SET(ctx, res, uint64);
+          }
+        }
+        if (end != *rest) {
+          if (rest_char == '.' || rest_char == 'e' || rest_char == 'E' ||
+              (rest_char >= '0' && rest_char <= '9')) {
+            char *rest_float;
+            double d;
+            errno = 0;
+            d = strtod(p, &rest_float);
+            if (!errno && rest_float == end) {
+              grn_obj_reinit(ctx, res, GRN_DB_FLOAT, 0);
+              GRN_FLOAT_SET(ctx, res, d);
+              *rest = rest_float;
+            } else {
+              return GRN_INVALID_ARGUMENT;
+            }
+          }
         }
       }
     }
@@ -2845,62 +2861,100 @@ grn_text_otoj(grn_ctx *ctx, grn_obj *bulk, grn_obj *obj, grn_obj_format *format)
     break;
   case GRN_UVECTOR :
     if (format) {
-      int i, j;
-      grn_id *v = (grn_id *)GRN_BULK_HEAD(obj), *ve = (grn_id *)GRN_BULK_CURR(obj);
-      int ncolumns = GRN_BULK_VSIZE(&format->columns) / sizeof(grn_obj *);
-      grn_obj **columns = (grn_obj **)GRN_BULK_HEAD(&format->columns);
-      GRN_TEXT_PUTS(ctx, bulk, "[[");
-      grn_text_itoa(ctx, bulk, ve - v);
-      GRN_TEXT_PUTC(ctx, bulk, ']');
-      if (v < ve) {
-        if (format->flags & GRN_OBJ_FORMAT_WITH_COLUMN_NAMES) {
-          GRN_TEXT_PUTS(ctx, bulk, ",[");
-          for (j = 0; j < ncolumns; j++) {
-            grn_id range_id;
-            if (j) { GRN_TEXT_PUTC(ctx, bulk, ','); }
-            GRN_TEXT_PUTS(ctx, bulk, "[");
-            GRN_BULK_REWIND(&buf);
-            grn_column_name_(ctx, columns[j], &buf);
-            grn_text_otoj(ctx, bulk, &buf, NULL);
-            GRN_TEXT_PUTC(ctx, bulk, ',');
-            /* column range */
-            range_id = grn_obj_get_range(ctx, columns[j]);
-            if (range_id == GRN_ID_NIL) {
-              GRN_TEXT_PUTS(ctx, bulk, "null");
-            } else {
-              int name_len;
-              grn_obj *range_obj;
-              char name_buf[GRN_TABLE_MAX_KEY_SIZE];
+      if (format->flags & GRN_OBJ_FORMAT_WITH_WEIGHT) {
+        int i, n;
+        grn_obj *domain;
 
-              range_obj = grn_ctx_at(ctx, range_id);
-              name_len = grn_obj_name(ctx, range_obj, name_buf,
-                                      GRN_TABLE_MAX_KEY_SIZE);
+        n = grn_uvector_size(ctx, obj);
+        domain = grn_ctx_at(ctx, obj->header.domain);
+        GRN_TEXT_PUTS(ctx, bulk, "{");
+        for (i = 0; i < n; i++) {
+          grn_id id;
+          unsigned int weight;
+
+          if (i > 0) {
+            GRN_TEXT_PUTC(ctx, bulk, ',');
+          }
+          id = grn_uvector_get_element(ctx, obj, i, &weight);
+          if (domain) {
+            if (domain->header.type == GRN_TABLE_NO_KEY) {
+              GRN_TEXT_PUTC(ctx, bulk, '"');
+              grn_text_ulltoa(ctx, bulk, id);
+              GRN_TEXT_PUTC(ctx, bulk, '"');
+            } else {
               GRN_BULK_REWIND(&buf);
-              GRN_TEXT_PUT(ctx, &buf, name_buf, name_len);
+              grn_table_get_key2(ctx, domain, id, &buf);
               grn_text_otoj(ctx, bulk, &buf, NULL);
             }
-            GRN_TEXT_PUTS(ctx, bulk, "]");
-          }
-          GRN_TEXT_PUTC(ctx, bulk, ']');
-        }
-        for (i = 0;; i++) {
-          GRN_TEXT_PUTS(ctx, bulk, ",[");
-          for (j = 0; j < ncolumns; j++) {
-            if (j) { GRN_TEXT_PUTC(ctx, bulk, ','); }
-            GRN_BULK_REWIND(&buf);
-            grn_obj_get_value(ctx, columns[j], *v, &buf);
-            grn_text_otoj(ctx, bulk, &buf, NULL);
-          }
-          GRN_TEXT_PUTC(ctx, bulk, ']');
-          v++;
-          if (v < ve) {
-            GRN_TEXT_PUTC(ctx, bulk, ',');
           } else {
-            break;
+            GRN_TEXT_PUTC(ctx, bulk, '"');
+            grn_text_ulltoa(ctx, bulk, id);
+            GRN_TEXT_PUTC(ctx, bulk, '"');
+          }
+          GRN_TEXT_PUTC(ctx, bulk, ':');
+          grn_text_ulltoa(ctx, bulk, weight);
+        }
+        GRN_TEXT_PUTS(ctx, bulk, "}");
+      } else {
+        /* TODO: Does we still need this code? If we don't need this, we should
+           remove this. */
+        int i, j;
+        grn_id *v = (grn_id *)GRN_BULK_HEAD(obj), *ve = (grn_id *)GRN_BULK_CURR(obj);
+        int ncolumns = GRN_BULK_VSIZE(&format->columns) / sizeof(grn_obj *);
+        grn_obj **columns = (grn_obj **)GRN_BULK_HEAD(&format->columns);
+        GRN_TEXT_PUTS(ctx, bulk, "[[");
+        grn_text_itoa(ctx, bulk, ve - v);
+        GRN_TEXT_PUTC(ctx, bulk, ']');
+        if (v < ve) {
+          if (format->flags & GRN_OBJ_FORMAT_WITH_COLUMN_NAMES) {
+            GRN_TEXT_PUTS(ctx, bulk, ",[");
+            for (j = 0; j < ncolumns; j++) {
+              grn_id range_id;
+              if (j) { GRN_TEXT_PUTC(ctx, bulk, ','); }
+              GRN_TEXT_PUTS(ctx, bulk, "[");
+              GRN_BULK_REWIND(&buf);
+              grn_column_name_(ctx, columns[j], &buf);
+              grn_text_otoj(ctx, bulk, &buf, NULL);
+              GRN_TEXT_PUTC(ctx, bulk, ',');
+              /* column range */
+              range_id = grn_obj_get_range(ctx, columns[j]);
+              if (range_id == GRN_ID_NIL) {
+                GRN_TEXT_PUTS(ctx, bulk, "null");
+              } else {
+                int name_len;
+                grn_obj *range_obj;
+                char name_buf[GRN_TABLE_MAX_KEY_SIZE];
+
+                range_obj = grn_ctx_at(ctx, range_id);
+                name_len = grn_obj_name(ctx, range_obj, name_buf,
+                                        GRN_TABLE_MAX_KEY_SIZE);
+                GRN_BULK_REWIND(&buf);
+                GRN_TEXT_PUT(ctx, &buf, name_buf, name_len);
+                grn_text_otoj(ctx, bulk, &buf, NULL);
+              }
+              GRN_TEXT_PUTS(ctx, bulk, "]");
+            }
+            GRN_TEXT_PUTC(ctx, bulk, ']');
+          }
+          for (i = 0;; i++) {
+            GRN_TEXT_PUTS(ctx, bulk, ",[");
+            for (j = 0; j < ncolumns; j++) {
+              if (j) { GRN_TEXT_PUTC(ctx, bulk, ','); }
+              GRN_BULK_REWIND(&buf);
+              grn_obj_get_value(ctx, columns[j], *v, &buf);
+              grn_text_otoj(ctx, bulk, &buf, NULL);
+            }
+            GRN_TEXT_PUTC(ctx, bulk, ']');
+            v++;
+            if (v < ve) {
+              GRN_TEXT_PUTC(ctx, bulk, ',');
+            } else {
+              break;
+            }
           }
         }
+        GRN_TEXT_PUTC(ctx, bulk, ']');
       }
-      GRN_TEXT_PUTC(ctx, bulk, ']');
     } else {
       grn_obj *range = grn_ctx_at(ctx, obj->header.domain);
       if (range && range->header.type == GRN_TYPE) {
@@ -2964,7 +3018,11 @@ grn_text_otoj(grn_ctx *ctx, grn_obj *bulk, grn_obj *obj, grn_obj_format *format)
       GRN_UINT32_INIT(&weight, 0);
       with_weight = (format && format->flags & GRN_OBJ_FORMAT_WITH_WEIGHT);
       n = grn_vector_size(ctx, obj);
-      GRN_TEXT_PUTC(ctx, bulk, '[');
+      if (with_weight) {
+        GRN_TEXT_PUTC(ctx, bulk, '{');
+      } else {
+        GRN_TEXT_PUTC(ctx, bulk, '[');
+      }
       for (i = 0; i < n; i++) {
         const char *_value;
         unsigned int _weight, length;
@@ -2978,19 +3036,19 @@ grn_text_otoj(grn_ctx *ctx, grn_obj *bulk, grn_obj *obj, grn_obj_format *format)
         } else {
           grn_obj_reinit(ctx, &value, obj->header.domain, 0);
         }
-        if (with_weight) {
-          GRN_TEXT_PUTC(ctx, bulk, '{');
-        }
         grn_bulk_write(ctx, &value, _value, length);
         grn_text_otoj(ctx, bulk, &value, NULL);
         if (with_weight) {
           GRN_TEXT_PUTC(ctx, bulk, ':');
           GRN_UINT32_SET(ctx, &weight, _weight);
           grn_text_otoj(ctx, bulk, &weight, NULL);
-          GRN_TEXT_PUTC(ctx, bulk, '}');
         }
       }
-      GRN_TEXT_PUTC(ctx, bulk, ']');
+      if (with_weight) {
+        GRN_TEXT_PUTC(ctx, bulk, '}');
+      } else {
+        GRN_TEXT_PUTC(ctx, bulk, ']');
+      }
       GRN_OBJ_FIN(ctx, &value);
       GRN_OBJ_FIN(ctx, &weight);
     }
